C***********************************************************************
C	FINITE ELEMENT ANALYSIS (ELASTIC)
C	TRIANGLE (2-D) ELEMENTS
C	TOMONAGA OKABE, 2009/11/21
C***********************************************************************
	PROGRAM TRIANGLE_2D_ELASTIC
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- INPUT FILE
	OPEN(1, FILE = 'INP_TRIANGLE.TXT')
C---- OUTPUT FILE
      OPEN(2, FILE = 'RST.CSV')
C-----------------------------------------------------------------------
	CALL INPUT
	DO 10 LELEM=1,NELEM
	CALL DMAT(LELEM)
	CALL BMAT(LELEM)
	CALL KMAT_LOCAL(LELEM)
	CALL MERGEKMAT(LELEM)
10	CONTINUE
	CALL BOUND
	CALL PGUSJO(TK,MAXDOF,NDOF)
	P(1:NDOF)=TK(1:NDOF,NDOF+1)
	CALL STRESS
	CALL CPU_TIME(TIME)
	WRITE(*,*) 'CPU TIME = ',TIME
	STOP
	END
c***********************************************************************
c	READ INPUT DATA FROM 'INP_TRIANGLE.TXT'
c***********************************************************************
	SUBROUTINE INPUT
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- MATERIAL PROPERTIES, NUMBER OF NODES, NUMBER OF ELEMENTS
	READ(1,*) EM,VNU,THICK
	READ(1,*) NNODE,NELEM
	NDOF = 2 * NNODE
C---- NUMBER OF BOUNDARY CONDITIONS
	READ(1,*) N_BC_DISP(1:2),N_BC_LOAD(1:2)
	N_BC_GIVEN = SUM(N_BC_DISP(1:2))
	N_LOAD_GIVEN=SUM(N_BC_LOAD(1:2))
C---- COORDINATES
	DO 10 INODE=1,NNODE
	READ(1,*) X(INODE),Y(INODE)
10	CONTINUE
C---- CONNECTIVITY
	DO 20 I=1,NELEM
	READ(1,*) LNODS(1:3,I)
20	CONTINUE
C---- BOUNDARY CONDITIONS OF DISPLACEMENTS
	J = 0
	DO 30 IDOF=1,2
	DO 40 I=1,N_BC_DISP(IDOF)
	J = J + 1
	READ(1,*)I_BC_GIVEN(J),V_BC_GIVEN(J)
	I_BC_GIVEN(J) = (I_BC_GIVEN(J) - 1) * 2 + IDOF
40	CONTINUE
30	CONTINUE
C---- BOUNDARY CONDITIONS OF LOADS
	P(1:NDOF) = 0.D0
	J=0
	DO 50 IDOF=1,2
	DO 60 I=1,N_BC_LOAD(IDOF)
	J = J + 1
	READ(1,*)I_LOAD_GIVEN(J),V_LOAD_GIVEN(J)
	I_LOAD_GIVEN(J) = (I_LOAD_GIVEN(J) - 1) * 2 + idof
	P(I_LOAD_GIVEN(J))=V_LOAD_GIVEN(J)
60	CONTINUE
50	CONTINUE
	RETURN
	END
C***********************************************************************
C	CONSTITUTIVE MATRIX:[D]
C***********************************************************************
	SUBROUTINE DMAT(LELEM)
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
	CC=EM/(1.D0-VNU*VNU)
	G=EM/2.D0/(1.D0+VNU)
	D(1,1,LELEM)=CC
	D(2,1,LELEM)=CC*VNU
	D(3,1,LELEM)=0.D0
	D(1,2,LELEM)=CC*VNU
	D(2,2,LELEM)=CC
	D(3,2,LELEM)=0.D0
	D(1,3,LELEM)=0.D0
	D(2,3,LELEM)=0.D0
	D(3,3,LELEM)=G
	RETURN
	END
C***********************************************************************
C	DISPLACEMENT-STRAIN MATRIX:[B] 
C***********************************************************************
	SUBROUTINE BMAT(LELEM)
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- LOCAL VARIABLES
	DIMENSION IP(3)
C---- AREA_2(2*AREA)
	DO 10 I=1,3
	IP(I)=LNODS(I,LELEM)
10	CONTINUE
	AREA_2=0.D0
	AREA_2=X(IP(2))*Y(IP(3))+X(IP(1))*Y(IP(2))+X(IP(3))*Y(IP(1))
	AREA_2=AREA_2-X(IP(2))*Y(IP(1))-X(IP(3))*Y(IP(2))-X(IP(1))*Y(IP(3))
	AREA(LELEM)=AREA_2/2.D0
C---- CONSTRUCT B MATRIX
	B(1:3,1:6,LELEM)=0.D0
	B(1,1,LELEM)=Y(IP(2))-Y(IP(3))
	B(1,3,LELEM)=Y(IP(3))-Y(IP(1))
	B(1,5,LELEM)=Y(IP(1))-Y(IP(2))
	B(2,2,LELEM)=X(IP(3))-X(IP(2))
	B(2,4,LELEM)=X(IP(1))-X(IP(3))
	B(2,6,LELEM)=X(IP(2))-X(IP(1))
	B(3,1,LELEM)=X(IP(3))-X(IP(2))
	B(3,2,LELEM)=Y(IP(2))-Y(IP(3))
	B(3,3,LELEM)=X(IP(1))-X(IP(3))
	B(3,4,LELEM)=Y(IP(3))-Y(IP(1))
	B(3,5,LELEM)=X(IP(2))-X(IP(1))
	B(3,6,LELEM)=Y(IP(1))-Y(IP(2))
	B(1:3,1:6,LELEM)=B(1:3,1:6,LELEM)/AREA_2
	RETURN
	END
C***********************************************************************
C	STIFFNESS MATRIX: [K]
C***********************************************************************
	SUBROUTINE KMAT_LOCAL(LELEM)
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- LOCAL VARIABLES
	DIMENSION A(3,6)
	A(1:3,1:6)=MATMUL(D(1:3,1:3,LELEM),B(1:3,1:6,LELEM))
	DK(1:6,1:6,LELEM)=MATMUL(TRANSPOSE(B(1:3,1:6,LELEM)),
	1A(1:3,1:6))
	DK(1:6,1:6,LELEM)=DK(1:6,1:6,LELEM)*AREA(LELEM)*THICK
	RETURN
	END
C***********************************************************************
C	TOTAL STIFFNESS MATRIX: [TK]
C***********************************************************************
	SUBROUTINE MERGEKMAT(LELEM)
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- LOCAL VARIABLES
	DIMENSION IP(6)
C---- ASSIGN THE NUMBER OF DOF IN THE TOTAL MATRIX
	DO 10 INODE =1,3
	DO 20 IDOF =1,2
	IP((INODE-1)*2+IDOF)=(LNODS(INODE,LELEM)-1)*2+IDOF
20	CONTINUE
10	CONTINUE
C---- MERGE DK INTO TK
	DO 30 I=1,6
	DO 40 J=1,6
	TK(IP(I),IP(J))=TK(IP(I),IP(J))+DK(I,J,LELEM)
40	CONTINUE
30	CONTINUE
	RETURN
	END
C***********************************************************************
C	BOUNDARY CONDITION
C***********************************************************************
	SUBROUTINE BOUND
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- DRAW THE EFFECT OF THE ROW OF I_BC_GIVEN FROM THE LOAD VECTOR
	DO 10 I=1,N_BC_GIVEN
	DO 20 K=1,NDOF
	P(K)=P(K)-TK(K,I_BC_GIVEN(I))*V_BC_GIVEN(I)
20	CONTINUE
10	CONTINUE
C---- MAKE THE I_BC_GIVEN ROW IN [TK] ZERO 
	DO 30 I=1,N_BC_GIVEN
	DO 40 K=1,NDOF
	TK(I_BC_GIVEN(I),K)=0.D0
40	CONTINUE
30	CONTINUE
C---- MAKE THE I_BC_GIVEN LINE IN [TK] ZERO 
	DO 50 I=1,N_BC_GIVEN
	DO 60 K=1,NDOF
	TK(K,I_BC_GIVEN(I))=0.D0
60	CONTINUE
50	CONTINUE
C---- MAKE THE DIAGONAL PARTS FOR THE I_BC_GIVEN
	DO 70 I=1,N_BC_GIVEN
	TK(I_BC_GIVEN(I),I_BC_GIVEN(I))=1.D0
70	CONTINUE
C---- FIX THE DISPLACEMENT BOUNDARY
	DO 80 I=1,N_BC_GIVEN
	P(I_BC_GIVEN(I))=V_BC_GIVEN(I)
80	CONTINUE

C---- INPUT THE LOAD VECTOR INTO TK
	DO 90 I=1,NDOF
	TK(I,NDOF+1)=P(I)
90	CONTINUE
	RETURN
	END
C***********************************************************************
C	STRESS CALCULATION
C***********************************************************************
	SUBROUTINE STRESS
	IMPLICIT REAL*8(A-H,O-Z)
	PARAMETER(
     1MAXNODE=150,
     2MAXDOF=300,
     3MAXMELEM=100,
     4MAXBC=100)
C---- MATERIAL PROPERTIES
	COMMON/MATERIAL/
     1EM,VNU
C---- VARIABLES FOR MODEL
	COMMON/MODEL/
     1X(MAXNODE),Y(MAXNODE),THICK,AREA(MAXMELEM),
     2NNODE,NDOF,NELEM,LNODS(3,MAXMELEM)
C---- MATRIX AND VECTOR
	COMMON/MATRIX/ 
     1P(MAXDOF),TK(MAXDOF,MAXDOF),DK(6,6,MAXMELEM),
     2B(3,6,MAXMELEM),D(3,3,MAXMELEM)
C---- BOUNDARY CONDITIONS
	COMMON/BOUNDARY/ 
     1V_BC_GIVEN(MAXBC),V_LOAD_GIVEN(MAXBC),
     2N_BC_GIVEN,N_LOAD_GIVEN,
     3I_BC_GIVEN(MAXBC),I_LOAD_GIVEN(MAXBC),
	4N_BC_DISP(MAXBC),N_BC_LOAD(MAXBC)
C---- RESULTS OF STRESS AND STRAIN
	COMMON/RESULTS/
     1STRN(3,MAXNODE),SIGM(3,MAXNODE)
C---- LOCAL VARIABLES
	DIMENSION DB(3,6),ST(3),SI(3),IP(6),U(6)
	DO 10 LELEM=1,NELEM
	DO 20 INODE =1,3
	DO 30 IDOF =1,2
	IP((INODE-1)*2+IDOF)=(LNODS(INODE,LELEM)-1)*2+IDOF
	U((INODE-1)*2+IDOF)=P(IP((INODE-1)*2+IDOF))
30	CONTINUE
20	CONTINUE
	ST(1:3)=MATMUL(B(1:3,1:6,LELEM),U(1:6))
	DB(1:3,1:6)=MATMUL(D(1:3,1:3,LELEM),B(1:3,1:6,LELEM))
	SI(1:3)=MATMUL(DB(1:3,1:6),U(1:6))
	DO 40 K=1,3
	STRN(1:3,LNODS(K,LELEM))=ST(1:3)
	SIGM(1:3,LNODS(K,LELEM))=SI(1:3)
40	CONTINUE
10	CONTINUE
	DO I=1,NNODE
	WRITE(2,1000)SIGM(1:3,I),STRN(1:3,I)
	ENDDO
1000	FORMAT(6(F20.10,','))
	RETURN
	END
C***********************************************************************
C    SIMULTANEOUS EQUATIONS SUBROUTINE
C    COMPLETE PIVOTING GAUSS-JORDAN METHOD
C***********************************************************************
      subroutine PGUSJO(TK,MAXDOF,NDOF)   
	IMPLICIT REAL*8(A-H,O-Z)
	DIMENSION TK(MAXDOF,MAXDOF)
	dimension a(NDOF,NDOF+1),L(NDOF) !z̒`
	integer P,Q !P,Q𐮐^Œ`
	DO I=1,NDOF
	DO J=1,NDOF+1
	A(I,J)=TK(I,J)
	ENDDO
	ENDDO
	N=NDOF
	M=1
      IND = 0 !IND̏ݒ
      NM = N + M !gs̗NM̒`
      EPS = 1.D-6 !l0͈̔EPS̒`
      DO J=1,N !m1N܂Ŕԍt
        L(J) = J
      ENDDO
      D = 1.0 !s񎮂̏ݒ
      DO K=1,N !1sNs܂őΊp
        AW = ABS(A(K,K)) !ől̏ݒ
        P = K !ꊷs̏ݒ
        Q = K !ꊷ̏ݒ
        DO J=K,N !k񂩂n܂ŐΒl̍őlT
          DO I=K,N !KsNs܂ŐΒl̍őlT
            IF(AW.LT.ABS(A(I,J)))THEN !AWΒlA(I,J)菬ꍇ
              AW = ABS(A(I,J)) !AWɐΒlA(I,J)
              P = I !ꊷsIɐݒ
              Q = J !ꊷJɐݒ
            ENDIF
          ENDDO
        ENDDO
        IF(AW.LT.EPS)THEN !ől0łꍇ
          IND = 11 !IND11
          RETURN !Tu[`̏I
        ELSE
	    IF(K.NE.P)THEN 
            D = -D !s񎮂̕ϊ
            DO J=K,NM !K񂩂NM܂ōsꊷ
              AW = A(K,J) !AWA(K,J)L
              A(K,J) = A(P,J)
              A(P,J) = AW
            ENDDO
          ENDIF
	    IF(K.NE.Q)THEN !K=P̏ꍇȂ
            D = -D !s񎮂̕ϊ
            DO I=1,N !1sNs܂ŗꊷ
              AW = A(I,K) !AWA(I,K)L
              A(I,K) = A(I,Q) !KQꊷ
              A(I,Q) = AW
            ENDDO
            MW = L(K)
            L(K) = L(Q) !m̔ԍ̕tւ
            L(Q) = MW
          ENDIF
          D = D*A(K,K) !s񎮂̌vZ
          KW = K + 1 !ZsKW̒`
          DO J=KW,NM !KW񂩂NM܂ŏ
            A(K,J) = A(K,J)/A(K,K) !Ksڂ̐K̉Z
          ENDDO
          DO I=1,N !1sNs܂ŏ
            IF(I.NE.K)THEN !Ίp͏Ȃ
              DO J=KW,NM !KW񂩂NM܂ŏ
                A(I,J) = A(I,J) - A(I,K)*A(K,J) !i|oj̉Z
              ENDDO
            ENDIF
          ENDDO
        ENDIF
      ENDDO
      NW = N + 1 !萔sB1ڂNWƂ
      DO J=NW,NM !萔sɏ
        DO I=1,N
          P = L(I) !isɑΉ関mԍpɑ
          A(P,N) = A(I,J) !mXבւĊgsANڂɊi[
        ENDDO
        DO I=1,N
          A(I,J) = A(I,N) !Nڂ萔sɈړ
        ENDDO
      ENDDO
	DO I=1,NDOF
	DO J=1,NDOF+1
	TK(I,J)=A(I,J)
	ENDDO
	ENDDO
      RETURN
      END